/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:expandtab:shiftwidth=2:tabstop=2:
 */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nsGNOMEShellSearchProvider.h"

#include "nsIBaseWindow.h"
#include "nsIDocShell.h"
#include "nsPIDOMWindow.h"
#include "mozilla/ModuleUtils.h"
#include "mozilla/Base64.h"
#include "nsIServiceManager.h"
#include "nsIWidget.h"
#include "nsIAppShellService.h"
#include "nsAppShellCID.h"
#include "nsPrintfCString.h"
#include "nsCOMPtr.h"
#include "nsGTKToolkit.h"
#include "nsINavHistoryService.h"
#include "nsToolkitCompsCID.h"
#include "nsIFaviconService.h"
#include "RemoteUtils.h"
#include "nsIStringBundle.h"
#include "imgIContainer.h"
#include "imgITools.h"
#include "nsNetCID.h"
#include "nsIIOService.h"

#include "mozilla/gfx/DataSurfaceHelpers.h"

#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>

using namespace mozilla;
using namespace mozilla::gfx;

#define KEYWORD_SEARCH_STRING "special:search"
#define KEYWORD_SEARCH_STRING_LEN 14

#define DBUS_BUS_NAME "org.mozilla.Firefox.SearchProvider"
#define DBUS_OBJECT_PATH "/org/mozilla/Firefox/SearchProvider"

static const char* introspect_template =
    "<!DOCTYPE node PUBLIC \"-//freedesktop//DTD D-BUS Object Introspection "
    "1.0//EN\"\n"
    "\"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd\";>\n"
    "<node>\n"
    " <interface name=\"org.freedesktop.DBus.Introspectable\">\n"
    "   <method name=\"Introspect\">\n"
    "     <arg name=\"data\" direction=\"out\" type=\"s\"/>\n"
    "   </method>\n"
    " </interface>\n"
    " <interface name=\"org.gnome.Shell.SearchProvider2\">\n"
    "   <method name=\"GetInitialResultSet\">\n"
    "     <arg type=\"as\" name=\"terms\" direction=\"in\" />\n"
    "     <arg type=\"as\" name=\"results\" direction=\"out\" />\n"
    "   </method>\n"
    "   <method name=\"GetSubsearchResultSet\">\n"
    "     <arg type=\"as\" name=\"previous_results\" direction=\"in\" />\n"
    "     <arg type=\"as\" name=\"terms\" direction=\"in\" />\n"
    "     <arg type=\"as\" name=\"results\" direction=\"out\" />\n"
    "   </method>\n"
    "   <method name=\"GetResultMetas\">\n"
    "     <arg type=\"as\" name=\"identifiers\" direction=\"in\" />\n"
    "     <arg type=\"aa{sv}\" name=\"metas\" direction=\"out\" />\n"
    "   </method>\n"
    "   <method name=\"ActivateResult\">\n"
    "     <arg type=\"s\" name=\"identifier\" direction=\"in\" />\n"
    "     <arg type=\"as\" name=\"terms\" direction=\"in\" />\n"
    "     <arg type=\"u\" name=\"timestamp\" direction=\"in\" />\n"
    "   </method>\n"
    "   <method name=\"LaunchSearch\">\n"
    "     <arg type=\"as\" name=\"terms\" direction=\"in\" />\n"
    "     <arg type=\"u\" name=\"timestamp\" direction=\"in\" />\n"
    "   </method>\n"
    "</interface>\n"
    "</node>\n";

DBusHandlerResult nsGNOMEShellSearchProvider::Introspect(DBusMessage* aMsg) {
  DBusMessage* reply;

  reply = dbus_message_new_method_return(aMsg);
  if (!reply) return DBUS_HANDLER_RESULT_NEED_MEMORY;

  dbus_message_append_args(reply, DBUS_TYPE_STRING, &introspect_template,
                           DBUS_TYPE_INVALID);

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

nsGNOMEShellSearchProvider::nsGNOMEShellSearchProvider()
    : mConnection(nullptr), mSearchSerial(0) {
  memset(mHistoryIcons, 0, sizeof(mHistoryIcons));
}

void nsGNOMEShellSearchProvider::SetHistoryIcon(int aSearchSerial,
                                                UniquePtr<uint8_t[]> aData,
                                                int aWidth, int aHeight,
                                                int aIconIndex) {
  MOZ_ASSERT(mSearchSerial == aSearchSerial);
  MOZ_ASSERT(aIconIndex < MAX_SEARCH_RESULTS_NUM);
  mHistoryIcons[aIconIndex].Set(aSearchSerial, std::move(aData), aWidth,
                                aHeight);
}

nsresult nsGNOMEShellSearchProvider::NewHistorySearch(const char* aSearchTerm) {
  // Initialize new search which invalidates all preview ones
  mSearchSerial++;

  nsresult rv;
  nsCOMPtr<nsINavHistoryQuery> histQuery;
  rv = mHistoryService->GetNewQuery(getter_AddRefs(histQuery));
  NS_ENSURE_SUCCESS(rv, rv);

  nsAutoCString searchTerm(aSearchTerm);
  rv = histQuery->SetSearchTerms(NS_ConvertUTF8toUTF16(searchTerm));
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsINavHistoryQueryOptions> histQueryOpts;
  rv = mHistoryService->GetNewQueryOptions(getter_AddRefs(histQueryOpts));
  NS_ENSURE_SUCCESS(rv, rv);

  // We want to get the URIs for every item in the user's history with the
  // given host
  rv = histQueryOpts->SetResultType(nsINavHistoryQueryOptions::RESULTS_AS_URI);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = histQueryOpts->SetSortingMode(
      nsINavHistoryQueryOptions::SORT_BY_VISITCOUNT_DESCENDING);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = histQueryOpts->SetMaxResults(MAX_SEARCH_RESULTS_NUM);
  NS_ENSURE_SUCCESS(rv, rv);

  // We only search history, because searching both bookmarks and history
  // is not supported, and history tends to be more comprehensive.
  rv = histQueryOpts->SetQueryType(
      nsINavHistoryQueryOptions::QUERY_TYPE_HISTORY);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsINavHistoryResult> histResult;
  rv = mHistoryService->ExecuteQuery(histQuery, histQueryOpts,
                                     getter_AddRefs(histResult));
  NS_ENSURE_SUCCESS(rv, rv);

  // Delete former search results
  mHistResultContainer = nullptr;

  rv = histResult->GetRoot(getter_AddRefs(mHistResultContainer));
  NS_ENSURE_SUCCESS(rv, rv);

  rv = mHistResultContainer->SetContainerOpen(true);
  NS_ENSURE_SUCCESS(rv, rv);

  uint32_t childCount = 0;
  rv = mHistResultContainer->GetChildCount(&childCount);
  NS_ENSURE_SUCCESS(rv, rv);

  return childCount != 0 ? NS_OK : NS_ERROR_FAILURE;
}

bool nsGNOMEShellSearchProvider::IsHistoryResultNodeURI(
    nsINavHistoryResultNode* aHistoryNode) {
  uint32_t type;
  nsresult rv = aHistoryNode->GetType(&type);
  if (NS_FAILED(rv) || type != nsINavHistoryResultNode::RESULT_TYPE_URI)
    return false;

  nsAutoCString title;
  rv = aHistoryNode->GetTitle(title);
  if (NS_SUCCEEDED(rv) && !title.IsEmpty()) {
    return true;
  }

  rv = aHistoryNode->GetUri(title);
  return NS_SUCCEEDED(rv) && !title.IsEmpty();
}

void nsGNOMEShellSearchProvider::GetIDKeyForURI(int aIndex, nsAutoCString& aUri,
                                                nsAutoCString& aIDKey) {
  // Compose ID as NN:URL where NN is index to our current history
  // result container.
  aIDKey = nsPrintfCString("%.2d:%s", aIndex, aUri.get());
}

int nsGNOMEShellSearchProvider::GetIndexFromIDKey(const char* aIDKey) {
  // ID is NN:URL where NN is index to our current history
  // result container.
  char tmp[] = {aIDKey[0], aIDKey[1], '\0'};
  return atoi(tmp);
}

class AsyncFaviconDataReady final : public nsIFaviconDataCallback {
 public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSIFAVICONDATACALLBACK

  AsyncFaviconDataReady(nsGNOMEShellSearchProvider* aSearchProvider,
                        int aIconIndex, int aSearchSerial)
      : mSearchProvider(aSearchProvider),
        mIconIndex(aIconIndex),
        mSearchSerial(aSearchSerial){};

 private:
  ~AsyncFaviconDataReady() {}

  nsGNOMEShellSearchProvider* mSearchProvider;
  int mIconIndex;
  int mSearchSerial;
};

NS_IMPL_ISUPPORTS(AsyncFaviconDataReady, nsIFaviconDataCallback)

// Inspired by SurfaceToPackedBGRA
static UniquePtr<uint8_t[]> SurfaceToPackedRGBA(DataSourceSurface* aSurface) {
  IntSize size = aSurface->GetSize();
  CheckedInt<size_t> bufferSize =
      CheckedInt<size_t>(size.width * 4) * CheckedInt<size_t>(size.height);
  if (!bufferSize.isValid()) {
    return nullptr;
  }
  UniquePtr<uint8_t[]> imageBuffer(new (std::nothrow)
                                       uint8_t[bufferSize.value()]);
  if (!imageBuffer) {
    return nullptr;
  }

  DataSourceSurface::MappedSurface map;
  if (!aSurface->Map(DataSourceSurface::MapType::READ, &map)) {
    return nullptr;
  }

  // Convert BGRA to RGBA
  uint32_t* aSrc = (uint32_t*)map.mData;
  uint32_t* aDst = (uint32_t*)imageBuffer.get();
  for (int i = 0; i < size.width * size.height; i++, aDst++, aSrc++) {
    *aDst = *aSrc & 0xff00ff00;
    *aDst |= (*aSrc & 0xff) << 16;
    *aDst |= (*aSrc & 0xff0000) >> 16;
  }

  aSurface->Unmap();

  return imageBuffer;
}

NS_IMETHODIMP
AsyncFaviconDataReady::OnComplete(nsIURI* aFaviconURI, uint32_t aDataLen,
                                  const uint8_t* aData,
                                  const nsACString& aMimeType,
                                  uint16_t aWidth) {
  // This is a callback from some previous search so we don't want it
  if (mSearchSerial != mSearchProvider->GetSearchSerial() || !aData ||
      !aDataLen) {
    return NS_ERROR_FAILURE;
  }

  // Decode the image from the format it was returned to us in (probably PNG)
  nsCOMPtr<imgIContainer> container;
  nsCOMPtr<imgITools> imgtool = do_CreateInstance("@mozilla.org/image/tools;1");
  nsresult rv = imgtool->DecodeImageFromBuffer(
      reinterpret_cast<const char*>(aData), aDataLen, aMimeType,
      getter_AddRefs(container));
  NS_ENSURE_SUCCESS(rv, rv);

  RefPtr<SourceSurface> surface = container->GetFrame(
      imgIContainer::FRAME_FIRST,
      imgIContainer::FLAG_SYNC_DECODE | imgIContainer::FLAG_ASYNC_NOTIFY);

  if (!surface || surface->GetFormat() != SurfaceFormat::B8G8R8A8) {
    return NS_ERROR_FAILURE;
  }

  // Allocate a new buffer that we own.
  RefPtr<DataSourceSurface> dataSurface = surface->GetDataSurface();
  UniquePtr<uint8_t[]> data = SurfaceToPackedRGBA(dataSurface);
  if (!data) {
    return NS_ERROR_OUT_OF_MEMORY;
  }

  mSearchProvider->SetHistoryIcon(mSearchSerial, std::move(data),
                                  surface->GetSize().width,
                                  surface->GetSize().height, mIconIndex);
  return NS_OK;
}

void nsGNOMEShellSearchProvider::ComposeSearchResultReply(
    DBusMessage* reply, const char* aSearchTerm, bool aSearchOnly) {
  uint32_t childCount = 0;
  nsresult rv = NS_OK;

  if (!aSearchOnly) {
    nsresult rv = mHistResultContainer->GetChildCount(&childCount);
  }

  DBusMessageIter iter;
  dbus_message_iter_init_append(reply, &iter);
  DBusMessageIter iterArray;
  dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY, "s", &iterArray);

  if (NS_SUCCEEDED(rv) && childCount > 0) {
    // Obtain the favicon service and get the favicon for the specified page
    nsCOMPtr<nsIFaviconService> favIconSvc(
        do_GetService("@mozilla.org/browser/favicon-service;1"));
    nsCOMPtr<nsIIOService> ios(do_GetService(NS_IOSERVICE_CONTRACTID));

    if (childCount > MAX_SEARCH_RESULTS_NUM) {
      childCount = MAX_SEARCH_RESULTS_NUM;
    }

    for (uint32_t i = 0; i < childCount; i++) {
      nsCOMPtr<nsINavHistoryResultNode> child;
      mHistResultContainer->GetChild(i, getter_AddRefs(child));
      if (NS_WARN_IF(NS_FAILED(rv))) {
        continue;
      }
      if (!child || !IsHistoryResultNodeURI(child)) {
        continue;
      }

      nsAutoCString uri;
      child->GetUri(uri);

      nsCOMPtr<nsIURI> iconIri;
      ios->NewURI(uri, nullptr, nullptr, getter_AddRefs(iconIri));
      nsCOMPtr<nsIFaviconDataCallback> callback =
          new AsyncFaviconDataReady(this, i, mSearchSerial);
      favIconSvc->GetFaviconDataForPage(iconIri, callback, 0);

      nsAutoCString idKey;
      GetIDKeyForURI(i, uri, idKey);

      const char* id = idKey.get();
      dbus_message_iter_append_basic(&iterArray, DBUS_TYPE_STRING, &id);
    }
  }

  nsPrintfCString searchString("%s:%s", KEYWORD_SEARCH_STRING, aSearchTerm);
  const char* search = searchString.get();
  dbus_message_iter_append_basic(&iterArray, DBUS_TYPE_STRING, &search);

  dbus_message_iter_close_container(&iter, &iterArray);
}

DBusHandlerResult nsGNOMEShellSearchProvider::GetInitialResultSet(
    DBusMessage* aMsg) {
  DBusMessage* reply;
  char** stringArray = nullptr;
  int elements;

  if (!dbus_message_get_args(aMsg, nullptr, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
                             &stringArray, &elements, DBUS_TYPE_INVALID) ||
      elements == 0) {
    reply = dbus_message_new_error(aMsg, DBUS_BUS_NAME, "Wrong argument");
  } else {
    reply = dbus_message_new_method_return(aMsg);
    nsresult rv = NewHistorySearch(stringArray[0]);
    ComposeSearchResultReply(reply, stringArray[0],
                             /* search only */ NS_FAILED(rv));
  }
  if (stringArray) {
    dbus_free_string_array(stringArray);
  }

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

DBusHandlerResult nsGNOMEShellSearchProvider::GetSubsearchResultSet(
    DBusMessage* aMsg) {
  DBusMessage* reply;

  char **unusedArray = nullptr, **stringArray = nullptr;
  int unusedNum, elements;

  if (!dbus_message_get_args(aMsg, nullptr, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
                             &unusedArray, &unusedNum, DBUS_TYPE_ARRAY,
                             DBUS_TYPE_STRING, &stringArray, &elements,
                             DBUS_TYPE_INVALID) ||
      elements == 0) {
    reply = dbus_message_new_error(aMsg, DBUS_BUS_NAME, "Wrong argument");
  } else {
    reply = dbus_message_new_method_return(aMsg);
    nsresult rv = NewHistorySearch(stringArray[0]);
    ComposeSearchResultReply(reply, stringArray[0],
                             /* search only */ NS_FAILED(rv));
  }

  if (unusedArray) {
    dbus_free_string_array(unusedArray);
  }
  if (stringArray) {
    dbus_free_string_array(stringArray);
  }

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

static void appendStringDictionary(DBusMessageIter* aIter, const char* aKey,
                                   const char* aValue) {
  DBusMessageIter iterDict, iterVar;
  dbus_message_iter_open_container(aIter, DBUS_TYPE_DICT_ENTRY, nullptr,
                                   &iterDict);
  dbus_message_iter_append_basic(&iterDict, DBUS_TYPE_STRING, &aKey);
  dbus_message_iter_open_container(&iterDict, DBUS_TYPE_VARIANT, "s", &iterVar);
  dbus_message_iter_append_basic(&iterVar, DBUS_TYPE_STRING, &aValue);
  dbus_message_iter_close_container(&iterDict, &iterVar);
  dbus_message_iter_close_container(aIter, &iterDict);
}

/*
  "icon-data": a tuple of type (iiibiiay) describing a pixbuf with width,
              height, rowstride, has-alpha,
              bits-per-sample, channels,
              image data
*/
void GnomeHistoryIcon::AppendIcon(DBusMessageIter* aIter) {
  DBusMessageIter iterDict, iterVar, iterStruct;
  dbus_message_iter_open_container(aIter, DBUS_TYPE_DICT_ENTRY, nullptr,
                                   &iterDict);
  const char* key = "icon-data";
  dbus_message_iter_append_basic(&iterDict, DBUS_TYPE_STRING, &key);
  dbus_message_iter_open_container(&iterDict, DBUS_TYPE_VARIANT, "(iiibiiay)",
                                   &iterVar);
  dbus_message_iter_open_container(&iterVar, DBUS_TYPE_STRUCT, nullptr,
                                   &iterStruct);

  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_INT32, &mWidth);
  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_INT32, &mHeight);
  int rowstride = mWidth * 4;
  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_INT32, &rowstride);
  int hasAlpha = true;
  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_BOOLEAN, &hasAlpha);
  int bitsPerSample = 8;
  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_INT32, &bitsPerSample);
  int channels = 4;
  dbus_message_iter_append_basic(&iterStruct, DBUS_TYPE_INT32, &channels);

  DBusMessageIter iterArray;
  dbus_message_iter_open_container(&iterStruct, DBUS_TYPE_ARRAY, "y",
                                   &iterArray);
  unsigned char* array = mData.get();
  dbus_message_iter_append_fixed_array(&iterArray, DBUS_TYPE_BYTE, &array,
                                       mWidth * mHeight * 4);
  dbus_message_iter_close_container(&iterStruct, &iterArray);

  dbus_message_iter_close_container(&iterVar, &iterStruct);
  dbus_message_iter_close_container(&iterDict, &iterVar);
  dbus_message_iter_close_container(aIter, &iterDict);
}

/* We can return those fields at GetResultMetas:
  "id": the result ID
  "name": the display name for the result
  "icon": a serialized GIcon (see g_icon_serialize()), or alternatively,
  "gicon": a textual representation of a GIcon (see g_icon_to_string()),
           or alternativly,
  "icon-data": a tuple of type (iiibiiay) describing a pixbuf with width,
              height, rowstride, has-alpha, bits-per-sample, and image data
  "description": an optional short description (1-2 lines)
*/
void nsGNOMEShellSearchProvider::AppendResultID(DBusMessageIter* aIter,
                                                const char* aID) {
  int index = GetIndexFromIDKey(aID);
  nsCOMPtr<nsINavHistoryResultNode> child;
  mHistResultContainer->GetChild(index, getter_AddRefs(child));
  nsAutoCString title;
  nsAutoCString uri;
  child->GetTitle(title);
  child->GetUri(uri);

  const char* titleStr = !(title.IsEmpty()) ? title.get() : uri.get();
  const char* descStr = uri.get();

  appendStringDictionary(aIter, "id", aID);
  appendStringDictionary(aIter, "name", titleStr);
  appendStringDictionary(aIter, "description", descStr);
  if (mHistoryIcons[index].GetSearchSerial() == mSearchSerial) {
    mHistoryIcons[index].AppendIcon(aIter);
  } else {
    appendStringDictionary(aIter, "gicon", "text-html");
  }
}

void nsGNOMEShellSearchProvider::AppendSearchID(DBusMessageIter* aIter,
                                                const char* aID) {
  if (strlen(aID) < KEYWORD_SEARCH_STRING_LEN + 2) {
    return;
  }
  appendStringDictionary(aIter, "id", KEYWORD_SEARCH_STRING);
  mSearchTerm = nsAutoCStringN<32>(aID + KEYWORD_SEARCH_STRING_LEN + 1);
  nsPrintfCString searchString(mGnomeSearchTitle.get(), mSearchTerm.get());
  appendStringDictionary(aIter, "name", searchString.get());
  appendStringDictionary(aIter, "gicon", "org.mozilla.Firefox");
}

DBusHandlerResult nsGNOMEShellSearchProvider::GetResultMetas(
    DBusMessage* aMsg) {
  DBusMessage* reply;
  char** stringArray;
  int elements;

  if (!dbus_message_get_args(aMsg, nullptr, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
                             &stringArray, &elements, DBUS_TYPE_INVALID) ||
      elements == 0) {
    reply = dbus_message_new_error(aMsg, DBUS_BUS_NAME, "Wrong argument");
  } else {
    reply = dbus_message_new_method_return(aMsg);

    DBusMessageIter iter;
    dbus_message_iter_init_append(reply, &iter);
    DBusMessageIter iterArray;
    dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY, "a{sv}",
                                     &iterArray);

    DBusMessageIter iterArray2;
    for (int i = 0; i < elements; i++) {
      dbus_message_iter_open_container(&iterArray, DBUS_TYPE_ARRAY, "{sv}",
                                       &iterArray2);
      if (strncmp(stringArray[i], KEYWORD_SEARCH_STRING,
                  KEYWORD_SEARCH_STRING_LEN) == 0) {
        AppendSearchID(&iterArray2, stringArray[i]);
      } else {
        AppendResultID(&iterArray2, stringArray[i]);
      }
      dbus_message_iter_close_container(&iterArray, &iterArray2);
    }

    dbus_message_iter_close_container(&iter, &iterArray);
    dbus_free_string_array(stringArray);
  }

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

void nsGNOMEShellSearchProvider::LaunchWithID(const char* aID,
                                              uint32_t aTimeStamp) {
  char* commandLine = nullptr;
  int tmp;

  if (strncmp(aID, KEYWORD_SEARCH_STRING, KEYWORD_SEARCH_STRING_LEN) == 0) {
    nsPrintfCString searchString("search:%s", mSearchTerm.get());
    const char* urlList[2] = {"unused", searchString.get()};
    commandLine = ConstructCommandLine(2, (char**)urlList, 0, &tmp);
  } else {
    int keyIndex = atoi(aID);
    nsCOMPtr<nsINavHistoryResultNode> child;
    mHistResultContainer->GetChild(keyIndex, getter_AddRefs(child));

    if (!child) {
      return;
    }

    nsAutoCString uri;
    nsresult rv = child->GetUri(uri);
    if (NS_FAILED(rv)) {
      return;
    }

    const char* urlList[2] = {"unused", uri.get()};
    commandLine = ConstructCommandLine(2, (char**)urlList, 0, &tmp);
  }

  if (commandLine) {
    HandleCommandLine(commandLine, aTimeStamp);
    free(commandLine);
  }
}

void nsGNOMEShellSearchProvider::LaunchWithAllResults(uint32_t aTimeStamp) {
  uint32_t childCount = 0;
  nsresult rv = mHistResultContainer->GetChildCount(&childCount);
  if (NS_FAILED(rv) || childCount == 0) {
    return;
  }

  if (childCount > MAX_SEARCH_RESULTS_NUM) {
    childCount = MAX_SEARCH_RESULTS_NUM;
  }

  char** urlList = (char**)moz_xmalloc(sizeof(char*) * (childCount + 2));
  int urlListElements = 0;

  urlList[urlListElements++] = strdup("unused");

  for (uint32_t i = 0; i < childCount; i++) {
    nsCOMPtr<nsINavHistoryResultNode> child;
    mHistResultContainer->GetChild(i, getter_AddRefs(child));

    if (!IsHistoryResultNodeURI(child)) {
      continue;
    }

    nsAutoCString uri;
    nsresult rv = child->GetUri(uri);
    if (NS_FAILED(rv)) {
      continue;
    }
    urlList[urlListElements++] = strdup(uri.get());
  }

  nsPrintfCString searchString("search:%s", mSearchTerm.get());
  urlList[urlListElements++] = strdup(searchString.get());

  int tmp;
  char* commandLine = ConstructCommandLine(urlListElements, urlList, 0, &tmp);
  if (commandLine) {
    HandleCommandLine(commandLine, aTimeStamp);
    free(commandLine);
  }

  for (int i = 0; i < urlListElements; i++) {
    free(urlList[i]);
  }
  free(urlList);
}

DBusHandlerResult nsGNOMEShellSearchProvider::ActivateResult(
    DBusMessage* aMsg) {
  DBusMessage* reply;
  char* resultID;
  char** stringArray;
  int elements;
  uint32_t timestamp;

  if (!dbus_message_get_args(aMsg, nullptr, DBUS_TYPE_STRING, &resultID,
                             DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &stringArray,
                             &elements, DBUS_TYPE_UINT32, &timestamp,
                             DBUS_TYPE_INVALID) ||
      resultID == nullptr) {
    reply = dbus_message_new_error(aMsg, DBUS_BUS_NAME, "Wrong argument");
  } else {
    reply = dbus_message_new_method_return(aMsg);
    LaunchWithID(resultID, timestamp);
    dbus_free_string_array(stringArray);
  }

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

DBusHandlerResult nsGNOMEShellSearchProvider::LaunchSearch(DBusMessage* aMsg) {
  DBusMessage* reply;
  char** stringArray;
  int elements;
  uint32_t timestamp;

  if (!dbus_message_get_args(aMsg, nullptr, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
                             &stringArray, &elements, DBUS_TYPE_UINT32,
                             &timestamp, DBUS_TYPE_INVALID) ||
      elements == 0) {
    reply = dbus_message_new_error(aMsg, DBUS_BUS_NAME, "Wrong argument");
  } else {
    reply = dbus_message_new_method_return(aMsg);
    LaunchWithAllResults(timestamp);
    dbus_free_string_array(stringArray);
  }

  dbus_connection_send(mConnection, reply, nullptr);
  dbus_message_unref(reply);

  return DBUS_HANDLER_RESULT_HANDLED;
}

DBusHandlerResult nsGNOMEShellSearchProvider::HandleDBusMessage(
    DBusConnection* aConnection, DBusMessage* aMsg) {
  NS_ASSERTION(mConnection == aConnection, "Wrong D-Bus connection.");

  const char* method = dbus_message_get_member(aMsg);
  const char* iface = dbus_message_get_interface(aMsg);

  if ((strcmp("Introspect", method) == 0) &&
      (strcmp("org.freedesktop.DBus.Introspectable", iface) == 0)) {
    return Introspect(aMsg);
  }

  if (strcmp("org.gnome.Shell.SearchProvider2", iface) == 0) {
    if (strcmp("GetInitialResultSet", method) == 0) {
      return GetInitialResultSet(aMsg);
    }
    if (strcmp("GetSubsearchResultSet", method) == 0) {
      return GetSubsearchResultSet(aMsg);
    }
    if (strcmp("GetResultMetas", method) == 0) {
      return GetResultMetas(aMsg);
    }
    if (strcmp("ActivateResult", method) == 0) {
      return ActivateResult(aMsg);
    }
    if (strcmp("LaunchSearch", method) == 0) {
      return LaunchSearch(aMsg);
    }
  }

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

void nsGNOMEShellSearchProvider::UnregisterDBusInterface(
    DBusConnection* aConnection) {
  NS_ASSERTION(mConnection == aConnection, "Wrong D-Bus connection.");
  // Not implemented
}

static DBusHandlerResult message_handler(DBusConnection* conn,
                                         DBusMessage* aMsg, void* user_data) {
  auto interface = static_cast<nsGNOMEShellSearchProvider*>(user_data);
  return interface->HandleDBusMessage(conn, aMsg);
}

static void unregister(DBusConnection* conn, void* user_data) {
  auto interface = static_cast<nsGNOMEShellSearchProvider*>(user_data);
  interface->UnregisterDBusInterface(conn);
}

static DBusObjectPathVTable remoteHandlersTable = {
    .unregister_function = unregister,
    .message_function = message_handler,
};

nsresult nsGNOMEShellSearchProvider::Startup() {
  if (mConnection && dbus_connection_get_is_connected(mConnection)) {
    // We're already connected so we don't need to reconnect
    return NS_ERROR_ALREADY_INITIALIZED;
  }

  nsCOMPtr<nsIStringBundleService> sbs =
      do_GetService(NS_STRINGBUNDLE_CONTRACTID);
  if (NS_WARN_IF(!sbs)) {
    return NS_ERROR_FAILURE;
  }

  nsCOMPtr<nsIStringBundle> bundle;
  sbs->CreateBundle("chrome://browser/locale/browser.properties",
                    getter_AddRefs(bundle));
  if (NS_WARN_IF(!bundle)) {
    return NS_ERROR_FAILURE;
  }

  nsAutoString searchTitle;
  bundle->GetStringFromName("gnomeSearchProviderSearch", searchTitle);
  mGnomeSearchTitle = NS_ConvertUTF16toUTF8(searchTitle);

  mHistoryService = do_GetService(NS_NAVHISTORYSERVICE_CONTRACTID);
  if (!mHistoryService) {
    return NS_ERROR_FAILURE;
  }

  mConnection =
      already_AddRefed<DBusConnection>(dbus_bus_get(DBUS_BUS_SESSION, nullptr));
  if (!mConnection) {
    return NS_ERROR_FAILURE;
  }
  dbus_connection_set_exit_on_disconnect(mConnection, false);
  dbus_connection_setup_with_g_main(mConnection, nullptr);

  DBusError err;
  dbus_error_init(&err);
  dbus_bus_request_name(mConnection, DBUS_BUS_NAME, DBUS_NAME_FLAG_DO_NOT_QUEUE,
                        &err);
  // The interface is already owned - there is another application/profile
  // instance already running.
  if (dbus_error_is_set(&err)) {
    dbus_error_free(&err);
    mConnection = nullptr;
    return NS_ERROR_FAILURE;
  }

  if (!dbus_connection_register_object_path(mConnection, DBUS_OBJECT_PATH,
                                            &remoteHandlersTable, this)) {
    mConnection = nullptr;
    return NS_ERROR_FAILURE;
  }

  return NS_OK;
}

void nsGNOMEShellSearchProvider::Shutdown() {
  if (!mConnection) {
    return;
  }

  dbus_connection_unregister_object_path(mConnection, DBUS_OBJECT_PATH);

  // dbus_connection_unref() will be called by RefPtr here.
  mConnection = nullptr;
}
